@@ -485,9 +485,9 @@ MODULE_AUTHOR("Karsten Keil");
 #ifdef MODULE_LICENSE
 MODULE_LICENSE("GPL");
 #endif
-MODULE_PARM(debug, "1i");
-MODULE_PARM(protocol, MODULE_PARM_T);
-MODULE_PARM(layermask, MODULE_PARM_T);
+module_param(debug, uint, 0400);
+module_param_array(protocol, int, NULL, 0400);
+module_param_array(layermask, int, NULL, 0400);
 #endif
 
 static char SpeedfaxName[] = "Speedfax";
