@@ -730,8 +730,8 @@ static const struct scsi_host_template sas_host_template = {
 	/* .name is initialized */
 	.name = "",
 	.queuecommand = sas_queuecommand,
-	.eh_strategy_handler = sas_scsi_recover_host,
-	.eh_timed_out = sas_scsi_timed_out,
+	//.eh_strategy_handler = sas_scsi_recover_host,
+	//.eh_timed_out = sas_scsi_timed_out,
 	.slave_alloc = sas_slave_alloc,
 	.slave_configure = sas_slave_configure,
 	.slave_destroy = sas_slave_destroy,
