@@ -92,7 +92,7 @@
 #if defined(__x86_64__)
 	#define WLAN_CPU_FAMILY		WLAN_x86_64
 	#define WLAN_SYSARCH		WLAN_PCAT
-#elif defined(__i386__) || defined(__i486__) || defined(__i586__) || defined(__i686__) || defined(__ia64__)
+#elif defined(__i386__) || defined(__i486__) || defined(__i586__) || defined(__i686__)
 	#define WLAN_CPU_FAMILY		WLAN_Ix86
 	#define WLAN_SYSARCH		WLAN_PCAT
 #elif defined(__ppc__)
@@ -143,11 +143,13 @@
    as new ones are discovered, add them below. 
 */
 
+#if defined(WLAN_HOSTIF)
 #if ((WLAN_HOSTIF == WLAN_PCI) || (WLAN_HOSTIF == WLAN_PLX))
 #if ((WLAN_SYSARCH == WLAN_SKIFF) || (WLAN_SYSARCH == WLAN_PMAC) || (WLAN_SYSARCH == WLAN_SPARC))
 #define REVERSE_ENDIAN
 #endif
 #endif
+#endif
 
 /*=============================================================*/
 /*------ Bit settings -----------------------------------------*/
@@ -213,7 +215,11 @@ typedef int64_t		INT64;
 /*=============================================================*/
 /*------ Compiler Portability Macros --------------------------*/
 /*=============================================================*/
+#if 0
 #define __WLAN_ATTRIB_PACK__		__attribute__ ((packed))
+#else
+#define __WLAN_ATTRIB_PACK__
+#endif
 
 /*=============================================================*/
 /*------ OS Portability Macros --------------------------------*/
